



def escape(x):
  return x
  

def setup_statserver(ws):
  global pygal, cstyle, escape
  try: 
    import asyncio
    from mrhttp import app
    import pygal
    import psutil
    
    #import mrworkserver
    from mrhttp import escape
    import tenjin, os
    #from tenjin.helpers import *
    tenjin.set_template_encoding('utf-8')
    engine = tenjin.Engine(path=[os.path.dirname(__file__)+'/html'],escapefunc='escape',tostrfunc='str')
    from pygal.style import DarkStyle
    cstyle = DarkStyle(
      title_font_size=28,
      label_font_size=24,
      major_label_font_size=24
    )
  except Exception as e:
    print("Enabling stats collection requires the following modules which appear to be missing:")
    print("pip install pygal psutil tenjin mrhttp")
    print(e)
    exit(0)


  @app.route('/{}')
  async def test(r,name):
    if name == "time":
      data = ws.async_times
      title = 'Processing time (ms)'
    elif name == "cpu":
      data = ws.cpu
      title = 'CPU Utilization'
    elif name == "mem":
      data = ws.mem
      title = 'Memory Usage'
    elif name in ws.counts:
      data = ws.counts[name]["cnts"]
      title = ws.counts[name]["title"]
    else:
      raise r.NotFound()
    b = await ws.loop.run_in_executor( ws.procpool, blocks, title, data )
   
    r.response.headers["Content-Type"] = 'image/svg+xml'
    return b.decode("utf-8")

  @app.route('/')
  async def index(r):
    imgs = ["time","cpu","mem"]
    imgs.extend( ws.counts.keys() )
    return engine.render('stats.ten', {"imgs":imgs})

  try:
    server = app.start_server(host="0.0.0.0", port=7099, loop=ws.loop)
  except Exception as e:
    print(e)
  return asyncio.ensure_future(server, loop=ws.loop)

async def stats_timer(self):
  minute = 0
  while 1:
    await asyncio.sleep(60)
    self.mem.append( psutil.virtual_memory().percent )
    self.cpu.append( psutil.cpu_percent() )

    for k in self.counts:
      c = self.counts[k]
      c["cnts"].append( c["cnt"] )
      c["cnt"] = 0

    if len( self.async_times ) > 0:
      tot = 0
      num = 0
      for ms in self.async_times:
        tot += ms
        num += 1
      avg = tot/num
      self.async_times_1m.append(avg)
      #self.async_times = []
    minute += 1
    if minute == 1440:
      # TODO Store 10m times to disk
      pass


def blocks(title, data):
  try:
    line_chart = pygal.Line(style=cstyle,show_legend=False)
    line_chart.title = title
    line_chart.add('', data)
    return line_chart.render()
  except Exception as e:
    print("chart exception", title, data, str(e))
    return b''





import socket, os, asyncio, time, struct
from concurrent.futures import ProcessPoolExecutor
from io import BytesIO

from mrworkserver import Protocol, CWorkServer

# TODO Remove stat server - just have a get_stats cmd and let the client graph
#from mrworkserver import setup_statserver, stats_timer
from mrworkserver import stats_timer


class WorkServer(CWorkServer):
  def __init__(self, *, protocol_factory=None, gather=False, collect_stats=False, callback=None, fetch_callback=None):
    self._loop = asyncio.new_event_loop()
    asyncio.set_event_loop(self.loop)
    self._connections = set()
    self._protocol_factory = protocol_factory or Protocol
    self._gather = gather
    self.cb = callback
    self.fetchcb = fetch_callback
    self.setcb   = self.nop
    self.async_times = None
    self.collect_stats = False
    self.async_times_1m = []
    self.stats_task = None
    #self.webserver_task = None
    if collect_stats:
      self.collect_stats = True
      self.cdata = {}
      
      self.async_times = []
      self.mem = []
      self.cpu = []
      self.counts = {}
      self.procpool = ProcessPoolExecutor(2)

      #try:
        #self.webserver_task = setup_statserver(self)
      #except Exception as e:
        #print(e)

      #server = app.create_server(host="0.0.0.0", port=5000)
      #self.webserver_task = asyncio.ensure_future(server)



    self.on_start = None
    self.on_stop  = None
    #self.workq = asyncio.Queue(loop=self.loop)
    super(WorkServer,self).__init__(callback, gather)


  @property
  def loop(self):
    if not self._loop:
      self._loop = asyncio.new_event_loop()
    return self._loop

  #def dcConnect(servers):
    #for s in servers:
      #srv = Server( s[0], s[1] )
      #srv.r, srv.w = await asyncio.open_connection( s[0], s[1], loop=self._loop, limit=DEFAULT_BUFFER_SIZE, ssl=ssl)
      #self.servers.append(srv)

  # Fetch
  def prefetch(self, o):
    b = self.fetchcb(self, o)
    if b == None: return None
    if not isinstance(b, bytes): return ""
    b = b'\x02' + struct.pack("<I", len(b)) + b
    return b    

  def nop(self, o):
    pass

  # Set
  def preset(self, k, v):
    self.setcb(self, k, v)

  # Stats
  def count_title(self, name, title):
    if not name in self.counts:
      self.counts[name] = { "title":title, "cnt":0, "cnts":[] }
    self.counts[name]["title"] = title

  def count(self, name, cnt):
    if not name in self.counts:
      self.counts[name] = { "title":name, "cnt":0, "cnts":[] }
    self.counts[name]["cnt"] += cnt 


  def run(self, host='0.0.0.0', port=7100, *, cores=None, ssl=None):

    if not asyncio.iscoroutinefunction(self.cb):
      print("WorkServer.cb must be an async function")
      return;

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    os.set_inheritable(sock.fileno(), True)

    loop = self.loop
    #asyncio.set_event_loop(loop)
    if self.on_start:
      loop.run_until_complete( self.on_start(self) )

    if self.collect_stats: 
      self.stats_task = asyncio.ensure_future( stats_timer(self) )

    server_coro = loop.create_server( lambda: self._protocol_factory(self), sock=sock, ssl=ssl)
    server = loop.run_until_complete(server_coro)

    try:
      loop.run_forever()
    except KeyboardInterrupt:
      pass
    finally:
      if self.stats_task: self.stats_task.cancel()
      #if self.webserver_task: self.webserver_task.cancel()
      server.close()
      #loop = asyncio.get_event_loop()
      if self.on_stop:
        loop.run_until_complete( self.on_stop(self) )
      loop.run_until_complete(server.wait_closed())
      loop.close()



